<?php

namespace Drupal\cforge\Installer;

use Drupal\mcapi\Entity\Currency;
use Drupal\address\Entity\AddressFormat;
use Drupal\field\Entity\FieldConfig;
use Drupal\taxonomy\Entity\Term;
use Drupal\user\Entity\User;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Core\Render\Element;
use Drupal\Core\Url;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

ini_set('max_execution_time', 150);

/**
 * Configuration options for each flavour.
 */
class Setup extends FormBase {

  /**
   * Constructor.
   */
  public function __construct($module_installer, $account_switcher) {
    $this->installer = $module_installer;
    $this->accountSwitcher = $account_switcher;
    // Contains cforge_setup2_form() and cforge_setup2_form_submit().
    $flavour = $this->config('cforge.settings')->get('flavour');
    module_load_install($flavour);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('module_installer'),
      $container->get('account_switcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cforge_setup_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // This uses the machine name.
    $flavour = $this->config('cforge.settings')->get('flavour');
    if (!$flavour) {
      throw new \Exception('Cannot run set up without a flavour.');
    }

    $form_state->loadInclude('cforge', 'install');
    $form['#title'] = t('Setup @flavour', ['@flavour' => $flavour]);
    $country = strtoupper($this->config('system.date')->get('country.default'));
    //@todo inject this
    $options = \Drupal::service('address.country_repository')->getList();

    if (isset($options[$country])) {
      $form['country'] = [
        '#type' => 'value',
        '#value' => $country,
      ];
    }
    else {
      $form['country'] = [
        '#type' => 'select',
        '#title' => $this->t('Country'),
        '#options' => $options,
        '#required' => TRUE,
        '#size' => 12,
        '#weight' => 2,
      ];
    }

    $form['data_content'] = [
      '#title' => t('Prepare the site with'),
      '#type' => 'radios',
      '#options' => [
        'example' => t('Example data'),
        'import' => t('Import Tools'),
        'none' => t('Nothing'),
      ],
      '#weight' => 5,
      'example' => [
        '#disabled' => !system_rebuild_module_data()['devel_generate'],
      ],
      '#required' => TRUE,
    ];
    $demo_data_class = cforge_demo_data_class_name();
    $form['neighbourhoods'] = [
      '#title' => $this->t('Neighbourhoods'),
      '#description' => $this->t('Limited list of possible neighbourhoods, one per line, in order. Leave blank if you plan to import data'),
      '#type' => 'textarea',
      '#default_value' => implode("\r\n", $demo_data_class::$hoods),
      '#weight' => 8,
      '#required' => 1,
      '#states' => [
        'visible' => [
          ':input[name=data_content]' => ['value' => 'none'],
        ],
      ],
    ];

    $vocab_names = array_keys($demo_data_class::vocabs());
    $form['default_categories'] = array(
      '#title' => $this->t('Preload ad categories'),
      '#description' => $this->t('Can be modified later'),
      '#type' => 'radios',
      '#options' => [0 => $this->t('- None -')] + array_combine($vocab_names, $vocab_names),
      '#default_value' => key($vocab_names),
      '#weight' => 9,
      '#states' => [
        'visible' => [
          ':input[name=data_content]' => ['value' => 'none'],
        ],
      ],
    );

    $form['currency_type'] = [
      '#title' => $this->t('Main currency type'),
      '#type' => 'radios',
      '#options' => [
        'hours' => $this->t('Hours'),
        'other' => $this->t('Other'),
      ],
      '#weight' => 6,
      '#required' => TRUE,
    ];
    $form['currency_name'] = [
      '#title' => $this->t('Currency name'),
      '#description' => $this->t('Use the plural'),
      '#type' => 'textfield',
      '#placeholder' => $this->t('Units'),
      '#weight' => 8,
      '#states' => [
        'visible' => [
          ':input[name="currency_type"]' => ['value' => 'other'],
        ],
      ],
    ];

    // Mimic the form_alter functionality except that the hook function is in
    // the flavour's install file.
    module_load_install($flavour);

    $form['actions'] = [
      'submit' => [
        '#type' => 'submit',
        '#value' => 'submit',
      ],
      '#weight' => 20,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Finally set the addressfield to the country's addressFormat.
    $country_code = $form_state->getValue('country');
    $field = FieldConfig::load('user.user.address');
    $settings = $field->get('settings');
    $settings['available_countries'][$country_code] = $country_code;
    $field->set('settings', $settings)->save();
    modify_address_formats($settings['available_countries']);

    $demo_data_class = cforge_demo_data_class_name();
    $this->accountSwitcher->switchTo(User::load(1));

    if ($form_state->getValue('default_categories')) {
      $terms = \Drupal::entityTypeManager()
        ->getStorage('taxonomy_term')
        ->loadByProperties(['vid' => SMALLAD_CATEGORIES]);
      // Delete any existing terms in the vocab (one is installed by default)
      foreach ($terms as $term) {
        $term->delete();
      }
      $vocab = Vocabulary::load(SMALLAD_CATEGORIES);
      // Now import the terms into the vocab.
      $this->vocabImportRecursive(
        $vocab,
        $demo_data_class::vocabs()[$form_state->getValue('default_categories')]
      );
      // Thie hierarchy property.
      $vocab->save();
    }

    // Set up the currency by deleting the others.
    if ($c = Currency::load('veur')) {
      $c->delete();
    }
    if ($form_state->getValue('currency_type') == 'hours') {
      if ($c = Currency::load('cc')) {
        $c->delete();
      }
    }
    else {
      if ($c = Currency::load('hhrs')) {
        $c->delete();
      }
      $curr = Currency::load('cc');
      $curr->name = $form_state->getValue('currency_name') ?: 'Units';
      $symbol = strtoupper(substr($curr->name, 0, 1));
      $curr->format = ["<strong>$symbol</strong>", '0000'];
      $curr->save();
    }

    if ($form_state->getValue('data_content') == 'example') {
      // Enable example modules and settings
      // @todo inject this
      \Drupal::service('module_installer')->install(
        ['views_ui',
          'field_ui',
          'contextual',
          'config',
          'admin_toolbar_tools',
          'devel',
          'devel_generate',
        ]
      );
      $this->configFactory()
        ->getEditable('views.settings')
        ->set('ui.show.sql_query.enabled', 1)
        ->set('ui.always_live_preview', 1)
        ->set('ui.show.advanced_column', 1)
        ->set('ui.show.master_display', 1)
        ->save();
      $this->configFactory()->getEditable('cforge.settings')
        ->set('demo_mode', TRUE)
        ->save();
      $hoods = $demo_data_class::$hoods;
    }
    else {
      if ($form_state->getValue('data_content') == 'import') {
        \Drupal::service('module_installer')->install(['cforge_import']);
      }
      $hoods = array_unique(explode("\n", trim($form_state->getValue('neighbourhoods'))));
    }
    $demo_data_class::createNeighbourhoods($hoods);
    $this->accountSwitcher->switchBack();

    $form_state->setRedirectUrl(
      Url::fromUserInput($this->config('cforge.settings')->get('member_frontpage'))
    );

  }

  /**
   * Import the terms in a vocabulary or a level of a vocab.
   */
  public function vocabImportRecursive(Vocabulary $vocab, array $termArray, $parent = 0) {
    $w = $hierarchy = 0;
    foreach (Element::Children($termArray) as $name) {
      $props = [
        'name' => $name,
        'weight' => $w++,
        'vid' => $vocab->id(),
        'parent' => $parent,
        'format' => 'plain_text',
      ];
      $term = Term::Create($props);
      $term->save();
      if ($termArray['#children']) {
        $vocab->setHierarchy(TAXONOMY_HIERARCHY_SINGLE);
        $this->vocabImportRecursive($vocab, $termArray[$name], $term->tid);
      }
    }
  }

}

/**
 * Ensure that every address format has a neighbourhood field.
 *
 * Deletes unused formats.
 *
 * @param array $country_codes
 *   Two-letter codes of countries this site supports, key & value.
 */
function modify_address_formats(array $country_codes) {
  $formats = AddressFormat::loadMultiple();

  foreach ($formats as $id => $format) {
    if (substr($id, 0, 3) == 'CF-') {
      //these don't need altering.
      continue;
    }
    // If we have an altered version of the address, overwrite that country's
    // address with ours.
    $alt = 'CF-' . $id;
    if (isset($formats[$alt])) {
      foreach ($formats[$alt]->toArray() as $key => $value) {
        if ($key != 'countryCode' && $key != 'uuid') {
          $format->set($key, $value);
        }
      }
      $format->save();
    }
    if (in_array($id, $country_codes)) {
      $pattern = $format->getFormat();
      if (strpos($pattern, '%dependentLocality' )) {
        continue;
      }
      $pattern .= "\n%dependentLocality";
      drupal_set_message("Modifying $id address format: ".$pattern, 'warning');
      $format->setFormat($pattern)
        ->setDependentLocalityType('neighborhood')
        ->setrequiredFields(['dependentLocality'])
        ->save();
    }
    else {
      if ($id != 'ZZ') {// @see Drupal\address\AddressFormatStorage
        $format->delete();
      }
    }
  }

}


